/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

/*! \file array.h
  \brief core file: defines the MT::Array class */

#ifndef MT_array_h
#define MT_array_h

#ifndef _template //you can enforce C-type templates when predefining these:
#  define _template template<class T>
#  define _Array Array
#  define _ArrayT Array<T>
#  define _ArrayUint Array<uint>
#endif

#include <iostream>

#define FOR1D(x,i)   for(i=0;i<x.d0;i++)
#define FOR1D_DOWN(x,i)   for(i=x.d0;i--;)
#define FOR2D(x,i,j) for(i=0;i<x.d0;i++) for(j=0;j<x.d1;j++)
#define FOR3D(x,i,j,k) for(i=0;i<x.d0;i++) for(j=0;j<x.d1;j++) for(k=0;k<x.d2;k++)
#define FOR_ALL(x,i)   for(i=0;i<x.N;i++)
#define forAll(i,A)  for(i=A.p;i!=A.pstop;i++)

typedef unsigned char byte;
typedef unsigned int uint;


//===========================================================================
//
// Array class
//

namespace MT{
  /*!\brief Simple array container to store arbitrary-dimensional arrays
    (tensors); can buffer more memory than necessary for faster
    resize; enables non-const reference of subarrays; enables fast
    memove for elementary types; implements many standard
    array/matrix/tensor operations. Please see the fully public attributes at the
    bottom of this page -- everthing is meant to be perfectly
    transparent. Interfacing with ordinary C-buffers is simple,
    e.g. via \c Array::referTo (Cbuffer,size) and \c Array::p and \c
    Array::pp. Please see also the reference for the \ref array.h
    header, which contains lots of functions that can be applied on
    Arrays. */

  _template class _Array{   //template <class T> class Array
  public:
    typedef _ArrayT ArrayT; //typedef Array<T> ArrayT
        
    T *p;     //!< the pointer on the linear memory allocated
    uint N;   //!< number of elements
    uint nd;  //!< number of dimensions
    uint d0;  //!< 0th dim
    uint d1;  //!< 1st dim
    uint d2;  //!< 2nd dim
    uint *d;  //!< pointer to dimensions (for nd<3 points to d0)
    T *pstop; //!< end of memory (pstop is already out of the bound)
    uint M;   //!< size of allocated memory (may be greater than N)
    bool reference;//!< true if this refers to some external memory
    T **pp;   //!< C-style 2D pointer (only valid if created with \c Array::getCarray!)


    //options
    /*!\brief be careful!!!; this option will modify the \c
      Array::resizeCopy : instead of calling the copy operator= for
      each element, the memory is moved with the \c memmove command -
      works only if you know what you do ... [default is true for
      simple types (like double, int, etc) and false for unknown
      ones] */
    bool memMove;

    /*!\brief if flexiMem is true (which is default!) the resize method will
      (1) at the first call allocate the exact amount of memory, (2)
      at further calls of increasing memory allocate twice the memory
      needed or (3) at further calls of decreasing memory only free
      the memory if the new size is smaller than a fourth */
    bool flexiMem;

    //! stores the sizeof(T)
    static int sizeT;

    //@}

#if 1//? garbage
    _ArrayUint *sparse; //!< to allocate element, column and row index lists (1+d0+d1 lists)
    enum matrixType{ full, diag };
    matrixType mtype;
#endif

  private:
    static char memMoveInit;
    void init();

  public:
    //constructors
    _Array();
    _Array(const ArrayT& a);
    _Array(uint i);
    _Array(uint i,uint j);
    _Array(uint i,uint j,uint k);
    _Array(uint i,uint j,uint k,uint l);
    _Array(const ArrayT& a,uint i);
    _Array(const ArrayT& a,uint i,uint j);
    _Array(T* p,uint size);
    ~_Array();

    ArrayT& operator=(const T& v);
    ArrayT& operator=(const ArrayT& a);

    //resizing
    ArrayT& resize    (uint D0);
    ArrayT& resizeCopy(uint D0);
    ArrayT& reshape   (uint D0);
    ArrayT& resize    (uint D0,uint D1);
    ArrayT& resizeCopy(uint D0,uint D1);
    ArrayT& reshape   (uint D0,uint D1);
    ArrayT& resize    (uint D0,uint D1,uint D2);
    ArrayT& resizeCopy(uint D0,uint D1,uint D2);
    ArrayT& reshape   (uint D0,uint D1,uint D2);
    ArrayT& resize    (uint D0,uint D1,uint D2,uint D3);
    ArrayT& resize    (uint ND,uint *dim);
    ArrayT& resizeCopy(uint ND,uint *dim);
    ArrayT& reshape   (uint ND,uint *dim);
    ArrayT& resize    (const _ArrayUint &newD);
    ArrayT& resizeCopy(const _ArrayUint &newD);
    ArrayT& reshape   (const _ArrayUint &newD);
    ArrayT& resizeAs    (const ArrayT& a);
    ArrayT& resizeCopyAs(const ArrayT& a);
    ArrayT& reshapeAs   (const ArrayT& a);

    //initializing/setting entries
    void clear();
    void setText(const char* str);
    void setZero(byte zero=0);
    void setUni(const T& scalar,int d=-1);
    void setId(int d=-1);
    void setDiag(const T& scalar,int d=-1);
    void setDiag(const ArrayT& vector);
    void setBlockMatrix(const ArrayT& A,const ArrayT& B,const ArrayT& C,const ArrayT& D);
    void setBlockVector(const ArrayT& a,const ArrayT& b);
    void setMatrixBlock(const ArrayT& B,uint lo0,uint lo1);  
    void setVectorBlock(const ArrayT& B,uint lo);            
    void setStraightPerm(int n=-1);
    void setReversePerm(int n=-1);
    void setRandomPerm(int n=-1);
    void setCarray(T *buffer,uint D0);
    void setCarray(T **buffer,uint D0,uint D1);
    void referTo(const T *buffer,uint n);
    void referTo(const ArrayT& a);
    void referToSubRange(const ArrayT& a,uint i,int I);
    void referToSubDim(const ArrayT& a,uint dim);
    void referToSubDim(const ArrayT& a,uint i,uint j);
    //void redirect(const ArrayT& a,uint i); //? remove; hardly more efficient than referToSubDim
    void takeOver(ArrayT& a);                   //a becomes a reference on its previously owned memory!
    void setGrid(uint dim,T lo,T hi,uint steps);

    //access
    uint D(uint k) const;
    _ArrayUint getDim() const{ _ArrayUint dd; dd.referTo(d,nd); return dd; }
    T& elem(uint i) const;
    T& scalar() const;
    T& last() const;
    T& operator()(uint i) const;
    T& operator()(uint i,uint j) const;
    T& operator()(uint i,uint j,uint k) const;
    T& operator()(const _ArrayUint &I) const;
    ArrayT operator[](uint i) const;
    ArrayT subDim(uint i,uint j) const;
    ArrayT& operator()();
    ArrayT sub(uint i,int I) const;
    ArrayT sub(uint i,int I,uint j,int J) const;
    ArrayT sub(uint i,int I,uint j,int J,uint k,int K) const;
    ArrayT sub(uint i,int I,_ArrayUint cols) const;
    void getMatrixBlock(ArrayT& B,uint lo0,uint lo1) const;
    void getVectorBlock(ArrayT& B,uint lo) const;
    T** getCarray() const;
    void copyInto  (T *buffer) const;
    void copyInto2D(T **buffer) const;
    T& min() const;
    T& max() const;
    T absMax() const;
    void minmax(T& minVal, T& maxVal) const;
    uint minIndex() const;
    uint maxIndex() const;
    void maxIndex(uint& i,uint& j) const;
    void maxIndex(uint& i,uint& j,uint& k) const;
    int findValue(const T& x) const;
    bool containsDoubles() const; // TL 17.07.08; whether at least one object is contained more than once
    uint getMemsize() const;
    void getIndexTuple(_ArrayUint &I,uint i) const;

    //appending etc
    T& append();
    void append(const T& x);
    void append(const ArrayT& x);
    void append(const T *p,uint n);
    void replicate(uint copies);
    void insert(uint i,const T& x);
    void replace(uint i,uint n,const ArrayT& x);
    void remove(uint i,uint n=1);
    void removePerm(uint i);          //more efficient for sets, works also for non-memMove arrays
    void removeValue(const T& x);
    bool removeValueSafe(const T& x); //? same as if((i=findValue(x))!=-1) remove[Perm](i);
    void delRow(uint i);
    void delColumns(uint i,uint k=1);
    void insRows(uint i,uint k=1);
    void insColumns(uint i,uint k=1);
    void resizeDim(uint k,uint dk);
    void setAppend(const T& x); //? same as if(findValue(x)==-1) append(x)
    void setAppend(const ArrayT& x);
    T popLast(){ T x; x=elem(N-1); remove(N-1); return x; }

    //sorting etc
    typedef bool (*ElemCompare)(const T& a,const T& b);
    uint findInSorted(const T& x,ElemCompare comp);
    uint insertSorted(const T& x,ElemCompare comp);
    void removeValueInSorted(const T& x,ElemCompare comp);

    //the array as permutation
    void permute(uint i,uint j);
    void permute(const _ArrayUint& permutation);
    void permuteInv(const _ArrayUint& permutation);
    void permuteRandomly();
    void shiftPerm(int offset);

    //sparse matrices [to be moved outside]
    double sparsity();
    void makeSparse();

    //I/O
    void write(std::ostream& os=std::cout,const char *ELEMSEP=" ",const char *LINESEP="\n ",bool binary=false) const;
    void read(std::istream& is,bool binary=false);
    //void readOld(std::istream& is); //? garbage
    void writeDim(std::ostream& os=std::cout) const;
    void readDim(std::istream& is);
    void writeRaw(std::ostream& os) const;
    void readRaw(std::istream& is);
    void writeWithIndex(std::ostream& os=std::cout) const; //? call writeList ?
    const ArrayT& ioraw() const;
    //const ArrayT& ionoraw() const;
    
    //kind of private
    void resizeMEM(uint n,bool copy);
    void freeMEM();
    void resetD();
  };
}

#ifndef MT_array_t_cpp

//===========================================================================
//
//!@name standard types
//

typedef MT::Array<double> arr;
typedef MT::Array<double> doubleA;
typedef MT::Array<float>  floatA;
typedef MT::Array<uint>   uintA;
typedef MT::Array<int>    intA;
typedef MT::Array<char>   charA;
typedef MT::Array<byte>   byteA;
typedef MT::Array<bool>   boolA;


//===========================================================================
//
// non-template functions //? needs most cleaning
//

arr Identity(uint dim);
void makeSymmetric(arr& A);
void transpose(arr& A);
void SUS(const arr& p,uint n,uintA& s);
uint SUS(const arr& p);

namespace MT{
  //! use this to turn on Lapack routines [default true if MT_LAPACK is defined]
  extern bool useLapack;
}

uint svd(const arr& A,arr& U,arr& d,arr& V,bool sort=true);
void svd(const arr& A,arr& U,arr& V);

uint inverse(arr& invA,const arr& A);
arr  inverse(const arr& A);
uint inverse_SVD(arr& inverse,const arr& A);
void inverse_LU(arr& invX,const arr& X);
void inverse_SymPosDef(arr& invA,const arr& A);
void pseudoInverse(arr& invA,const arr& A,const arr& W,double robustnessEps);

double determinant(const arr& A);
double cofactor(const arr& A,uint i,uint j);

//void getIndexTuple(uintA &I,uint i,const uintA &d);  //? that also exists inside of array!
void lognormScale(arr& P,double& logP,bool force=true);

void gnuplot(const arr& X);
void write(const arr& X,const arr& Y,const char* name);
void write(const arr& X,const arr& Y,const arr& Z,const char* name);
void write(const arr& X,const arr& Y,const arr& Z,const arr& W,const char* name);

void write_ppm(const byteA &img,const char *file_name,bool swap_rows=false);
void read_ppm(byteA &img,const char *file_name,bool swap_rows=false);
void add_alpha_channel(byteA &img,byte alpha);
void make_grey(byteA &img);
void make_RGB(byteA &img);
void flip_image(byteA &img);

void scanArrFile(const char* name);


//===========================================================================
//
// template functions
//

_template void transpose(MT::_ArrayT& x,const MT::_ArrayT& y);
_template void getDiag(MT::_ArrayT& x,const MT::_ArrayT& y);
_template void inverse2d(MT::_ArrayT& invA,const MT::_ArrayT& A);

_template T entropy(const MT::_ArrayT& v);
_template T normalizeDist(MT::_ArrayT& v);
_template void makeConditional(MT::_ArrayT& P);
_template void checkNormalization(MT::_ArrayT& v,double tol);
_template void checkNormalization(MT::_ArrayT& v){ checkNormalization(v,1e-10); }
_template void eliminate(MT::_ArrayT& x,const MT::_ArrayT& y,uint d);
_template void eliminate(MT::_ArrayT& x,const MT::_ArrayT& y,uint d,uint e);
_template void eliminatePartial(MT::_ArrayT& x,const MT::_ArrayT& y,uint d);

#ifndef SWIG
_template T sqrDistance(const MT::_ArrayT& v, const MT::_ArrayT& w);
_template T maxDiff(const MT::_ArrayT& v,const MT::_ArrayT& w,uint *maxi=0);
_template T maxRelDiff(const MT::_ArrayT& v,const MT::_ArrayT& w,T tol);
//_template T sqrDistance(const MT::_ArrayT& v, const MT::_ArrayT& w,const MT::Array<bool>& mask);
_template T sqrDistance(const MT::_ArrayT& g,const MT::_ArrayT& v,const MT::_ArrayT& w);
_template T euclideanDistance(const MT::_ArrayT& v, const MT::_ArrayT& w);
_template T metricDistance(const MT::_ArrayT& g,const MT::_ArrayT& v, const MT::_ArrayT& w);

_template T sum(const MT::_ArrayT& v);
_template T sumOfAbs(const MT::_ArrayT& v);
_template T sumOfSqr(const MT::_ArrayT& v);
_template T norm(const MT::_ArrayT& v);
_template T mean(const MT::_ArrayT& v);
_template T product(const MT::_ArrayT& v);

_template T trace(const MT::_ArrayT& v);
_template T var(const MT::_ArrayT& v);
_template T minDiag(const MT::_ArrayT& v);

_template void innerProduct(MT::_ArrayT& x,const MT::_ArrayT& y, const MT::_ArrayT& z);
_template void outerProduct(MT::_ArrayT& x,const MT::_ArrayT& y, const MT::_ArrayT& z);
_template T scalarProduct(const MT::_ArrayT& v,const MT::_ArrayT& w);
_template T scalarProduct(const MT::_ArrayT& g,const MT::_ArrayT& v, const MT::_ArrayT& w);

    //tensor operations [to be moved out of class declaration]
_template void tensorCondNormalize(MT::_ArrayT &X,int left);
_template void tensorCondMax      (MT::_ArrayT &X,uint left);
_template void tensorCondSoftMax  (MT::_ArrayT &X,uint left,double beta);
_template void tensorCheckCondNormalization          (const MT::_ArrayT &X,uint left,double tol=1e-10);
_template void tensorCheckCondNormalization_with_logP(const MT::_ArrayT &X,uint left, double logP, double tol=1e-10);

_template void tensorEquation    (MT::_ArrayT &X,const MT::_ArrayT &A,const uintA &pickA,const MT::_ArrayT &B,const uintA &pickB,uint sum=0);
_template void tensorMarginal    (MT::_ArrayT &Y,const MT::_ArrayT &X,const uintA &Yid);
_template void tensorMaxMarginal (MT::_ArrayT &Y,const MT::_ArrayT &X,const uintA &Yid);
_template void tensorMarginal_old(MT::_ArrayT &y,const MT::_ArrayT &x,const uintA &xd,const uintA &ids);
_template void tensorMultiply    (MT::_ArrayT &X,const MT::_ArrayT &Y,const uintA &Yid);
_template void tensorAdd         (MT::_ArrayT &X,const MT::_ArrayT &Y,const uintA &Yid);
_template void tensorMultiply_old(MT::_ArrayT &x,const MT::_ArrayT &y,const uintA &d,const uintA &ids);
_template void tensorDivide      (MT::_ArrayT &X,const MT::_ArrayT &Y,const uintA &Yid);

//arrays interpreted as a set (container)
_template void setUnite(MT::_ArrayT& x,const MT::_ArrayT& y, const MT::_ArrayT& z);
_template void setSection(MT::_ArrayT& x,const MT::_ArrayT& y, const MT::_ArrayT& z);
_template void setMinus(MT::_ArrayT& x,const MT::_ArrayT& y);
_template uint numberSharedElements(const MT::_ArrayT& x,const MT::_ArrayT& y); //? sectionSize
//? initialization members ??
_template void rndInt(MT::_ArrayT& a,int low=0,int high=1,bool add=false);
_template void rndUni(MT::_ArrayT& a,double low=0.,double high=1.,bool add=false);
_template void rndGauss(MT::_ArrayT& a,double stdDev=1.,bool add=false);
//_template void rndGauss(MT::_ArrayT& a,bool add=false);
//_template MT::_ArrayT& rndGauss(double stdDev,uint dim);
_template uint softMax(const MT::_ArrayT& a,arr& soft,double beta);
#endif

_template void writeArr(const MT::_ArrayT& x,const char* tag,std::ofstream& os); //? ostream als erstes argument (call writeTagged -> als member?)
_template bool readArr(MT::_ArrayT& x,const char *tag,std::ifstream& is); //?istream als erstes argument


//===========================================================================

template<class T> MT::Array<T> TUPLE(const T& i){                                    MT::Array<T> z(1); z(0)=i; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j){                               MT::Array<T> z(2); z(0)=i; z(1)=j; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k){                          MT::Array<T> z(3); z(0)=i; z(1)=j; z(2)=k; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k,const T& l){                     MT::Array<T> z(4); z(0)=i; z(1)=j; z(2)=k; z(3)=l; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k,const T& l,const T& m){                MT::Array<T> z(5); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k,const T& l,const T& m,const T& n){           MT::Array<T> z(6); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k,const T& l,const T& m,const T& n,const T& o){      MT::Array<T> z(7); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; return z; }
template<class T> MT::Array<T> TUPLE(const T& i,const T& j,const T& k,const T& l,const T& m,const T& n,const T& o,const T& p){ MT::Array<T> z(8); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; z(7)=p; return z; }

#define TUP TUPLE<uint>
#define ARR TUPLE<double>
/*
inline uintA TUP(uint i){                                           uintA z(1); z(0)=i; return z; }
inline uintA TUP(uint i,uint j){                                    uintA z(2); z(0)=i; z(1)=j; return z; }
inline uintA TUP(uint i,uint j,uint k){                             uintA z(3); z(0)=i; z(1)=j; z(2)=k; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l){                      uintA z(4); z(0)=i; z(1)=j; z(2)=k; z(3)=l; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m){               uintA z(5); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n){        uintA z(6); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n,uint o){ uintA z(7); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; return z; }
inline uintA TUP(uint i,uint j,uint k,uint l,uint m,uint n,uint o,uint p){ uintA z(8); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; z(7)=p; return z; }

inline arr ARR(double i){                                                  arr z(1); z(0)=i; return z; }
inline arr ARR(double i,double j){                                        arr z(2); z(0)=i; z(1)=j; return z; }
inline arr ARR(double i,double j,double k){                             arr z(3); z(0)=i; z(1)=j; z(2)=k; return z; }
inline arr ARR(double i,double j,double k,double l){                      arr z(4); z(0)=i; z(1)=j; z(2)=k; z(3)=l; return z; }
inline arr ARR(double i,double j,double k,double l,double m){               arr z(5); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; return z; }
inline arr ARR(double i,double j,double k,double l,double m,double n){        arr z(6); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; return z; }
inline arr ARR(double i,double j,double k,double l,double m,double n,double o){ arr z(7); z(0)=i; z(1)=j; z(2)=k; z(3)=l; z(4)=m; z(5)=n; z(6)=o; return z; }
*/

//===========================================================================
//
//!@name basic operators
//

namespace MT{
  _template _ArrayT operator~(const _ArrayT& y);
  _template _ArrayT operator-(const _ArrayT& y);
  _template _ArrayT operator^(const _ArrayT& y,const _ArrayT& z);
  _template _ArrayT operator*(const _ArrayT& y,const _ArrayT& z);
  _template _ArrayT operator*(const _ArrayT& y,T z);
  _template _ArrayT operator*(T y,const _ArrayT& z);
  
#define BinaryOperator( op, name)					\
  _template _ArrayT operator op(const _ArrayT& y,const _ArrayT& z); \
  _template _ArrayT operator op(T y,const _ArrayT& z);	\
  _template _ArrayT operator op(const _ArrayT& y,T z);
  BinaryOperator( + , plusA);
  BinaryOperator( - , minusA);
  BinaryOperator( % , mult);
  BinaryOperator( / , div);
#undef BinaryOperator
  
#ifndef SWIG
#define CompoundAssignmentOperator( op )				\
  _template _ArrayT& operator op (_ArrayT& x,const _ArrayT& y); \
  _template _ArrayT& operator op ( _ArrayT& x, T y );
  CompoundAssignmentOperator( |= );
  CompoundAssignmentOperator( ^= );
  CompoundAssignmentOperator( &= );
  CompoundAssignmentOperator( += );
  CompoundAssignmentOperator( -= );
  CompoundAssignmentOperator( *= );
  CompoundAssignmentOperator( /= );
  CompoundAssignmentOperator( %= );
#undef CompoundAssignmentOperator
#endif
}

#define UnaryOperation( name, op )					\
  _template MT::_ArrayT& name (MT::_ArrayT& x,const MT::_ArrayT& y); 
UnaryOperation( negative, - )
#undef UnaryOperation

#define BinaryOperation( name, op )					\
  _template MT::_ArrayT& name(MT::_ArrayT& x,const MT::_ArrayT& y,const MT::_ArrayT& z); \
  _template MT::_ArrayT& name##S(MT::_ArrayT& x,const MT::_ArrayT& y,T z); \
  _template MT::_ArrayT& name##S(MT::_ArrayT& x, T y, const MT::_ArrayT& z);
BinaryOperation( plusA , + );
BinaryOperation( minusA , - );
BinaryOperation( mult , * );
BinaryOperation( div , / );
#undef BinaryOperation

#ifndef SWIG
#define UnaryFunction( func )						\
  _template MT::_ArrayT func (const MT::_ArrayT& y);
UnaryFunction( acos );
UnaryFunction( asin );
UnaryFunction( atan );
UnaryFunction( cos );
UnaryFunction( sin );
UnaryFunction( tan );
UnaryFunction( cosh );
UnaryFunction( sinh );
UnaryFunction( tanh );
UnaryFunction( acosh );
UnaryFunction( asinh );
UnaryFunction( atanh );
UnaryFunction( exp );
UnaryFunction( log );
UnaryFunction( log10 );
UnaryFunction( sqrt );
UnaryFunction( cbrt );
UnaryFunction( ceil );
UnaryFunction( fabs );
UnaryFunction( floor );
#undef UnaryFunction

#define BinaryFunction( func )						\
  _template MT::_ArrayT func(const MT::_ArrayT& y,const MT::_ArrayT& z); \
  _template MT::_ArrayT func(const MT::_ArrayT& y,T z);	\
_template MT::_ArrayT func(T y,const MT::_ArrayT& z);
BinaryFunction( atan2 );
BinaryFunction( pow );
BinaryFunction( fmod );
#undef BinaryFunction

_template std::istream& operator>>(std::istream& is,MT::_ArrayT& x);
_template MT::_ArrayT& operator<<(MT::_ArrayT& x,const char* str);
_template std::ostream& operator<<(std::ostream& os,const MT::_ArrayT& x);
_template void checkNan(const MT::_ArrayT& x);
_template bool operator==(const MT::_ArrayT& v,const MT::_ArrayT& w);
_template bool operator==(const MT::_ArrayT& v,const T *w);
_template bool operator!=(const MT::_ArrayT& v,const MT::_ArrayT& w);
_template bool operator<(const MT::_ArrayT& v,const MT::_ArrayT& w);
#endif

//===========================================================================
//
// double template functions
//

#ifndef SWIG
template<class T,class S> void resizeAs    (MT::Array<T>& x,const MT::Array<S>& a){
  x.nd=a.nd; x.d0=a.d0; x.d1=a.d1; x.d2=a.d2;
  x.resetD();
  if(x.nd>3){ x.d=new uint[x.nd];  memmove(x.d,a.d,x.nd*sizeof(uint)); }
  x.resizeMEM(a.N,false);
}
template<class T,class S> void resizeCopyAs(MT::Array<T>& x,const MT::Array<S>& a);
template<class T,class S> void reshapeAs   (MT::Array<T>& x,const MT::Array<S>& a);
template<class T,class S> void copy        (MT::Array<T>& x,const MT::Array<S>& a){
  resizeAs(x,a);
  for(uint i=0;i<x.N;i++) x.elem(i)=(T)a.elem(i);
}
//! check whether this and \c a have same dimensions
template<class T,class S>
bool samedim(const MT::Array<T>& a,const MT::Array<S>& b){
  return (b.nd==a.nd && b.d0==a.d0 && b.d1==a.d1 && b.d2==a.d2);
}
#endif


//===========================================================================
//
// lapack interfaces
//

void blas_Mv(arr& y,const arr& A,const arr& x);
void blas_MM(arr& X,const arr& A,const arr& B);
void blas_MsymMsym(arr& X,const arr& A,const arr& B);
uint lapack_SVD(const arr& A, arr& U, arr& d, arr& Vt);
void lapack_LU(const arr& A, arr& LU);
void lapack_EigenDecomp(const arr& symmA, arr& Evals, arr& Evecs);
bool lapack_isPositiveSemiDefinite(const arr& symmA);
void lapack_inverseSymPosDef(arr& invA,const arr& A);
void lapack_Ainv_b_sym(arr& x,const arr& A, const arr& b);


//===========================================================================
//
// implementations
//

#if defined MT_IMPLEMENT_TEMPLATES | defined MT_IMPLEMENTATION
#  include"array_t.cpp"
#endif

#ifdef MT_IMPLEMENTATION
#  include"array.cpp"
#  include"array_lapack.cpp"
#endif
               
#endif
#endif
