#define MT_std_t_cpp

#include"std.h"

namespace MT{
  /*!\brief a standard method to save an object into a file. The same as
    std::ofstream file; MT::open(file,filename); file <<x;
    file.close(); */
  template<class T> void save(const T& x,const char *filename){
    std::ofstream file;
    open(file,filename);
    file <<x;
    file.close();
  }

    /*!\brief a standard method to load object from a file. The same as
    std::ifstream file; MT::open(file,filename); file >>x;
    file.close(); */
  template<class T> void load(T& x,const char *filename){
    std::ifstream file;
    open(file,filename);
    file >>x;
    file.close();
  }

    /*!\brief Search for a command line option \c -tag and, if found, pipe the
     next command line option into \c value by the
     \c operator>>(istream&,type&). Returns false on failure. */
  template<class T>
  bool getFromCmdLine(T& x,const char *tag){
    char *opt=getCmdLineArgument(tag);
    if(!opt) return false;
    std::istringstream s(opt);
    s >>x;
    if(s.fail()) HALT("error when reading parameter from command line: " <<tag);
    return true;
  }

  /*!\brief Search the first occurence of a sequence '\c tag:'
  in the config file (opened automatically) and, if found, pipes
  it in \c value. Returns false if parameter is not found. */
  template<class T>
  bool getFromCfgFile(T& x,const char *tag){
    if(!cfgOpenFlag) openConfigFile();
    CHECK(!cfgLock,"cfg file is locked");
    cfgLock=true;
    cfgFile.clear();
    cfgFile.seekg(std::ios::beg);
    if(!cfgFile.good()){ cfgLock=false; return false; }
    
    unsigned n=strlen(tag);
    char *buf=new char [n+2]; memset(buf,0,n+2);
    while(cfgFile.good()){
      memmove(buf,buf+1,n);
      buf[n]=cfgFile.get();
      if(buf[n]==' ' || buf[n]=='\t' || buf[n]==':' || buf[n]=='='){ buf[n]=0; if(!strcmp(tag,buf)) break; buf[n]=':'; }
    };
    delete[] buf;
    
    if(!cfgFile.good()){ cfgLock=false; return false; }
    
    skip(cfgFile," :=\n\r\t");
    cfgFile >>x;
    
    if(cfgFile.fail()) HALT("error when reading parameter " <<tag);
    cfgLock=false;
    return true;
  }
  
  template<class T>
  void getParameterBase(T& x,const char *tag,bool hasDefault,const T* Default){
    log() <<std::setw(20) <<tag <<" = " <<std::setw(5);
    log().flush();
    
    if(getFromCmdLine(x,tag)){
      log() <<x <<" [" <<typeid(x).name() <<"] (cmd line!)" <<std::endl;
      return;
    }
    
    if(getFromCfgFile(x,tag)){
      log() <<x <<" [" <<typeid(x).name() <<"]" <<std::endl;
      return;
    }
    
    if(hasDefault){
      x=*Default;
      log() <<x <<" [" <<typeid(x).name() <<"] (default!)"<<std::endl;
      return;
    }
    
    HALT("could not initialize parameter `" <<tag
	     <<"': parameter has no default;\n     either use command option `-"
	     <<tag <<" ...' or specify `"
	     <<tag <<"= ...' in the config file `MT.cfg'");
  }

  template<class T> T getParameter(const char *tag){
    T x;
    getParameterBase<T>(x,tag,false,(T*)NULL);
    return x;
  }
  template<class T> T getParameter(const char *tag,const T& Default){
    T x;
    getParameterBase<T>(x,tag,true,&Default);
    return x;
  }
  template<class T> void getParameter(T& x,const char *tag,const T& Default){
    getParameterBase<T>(x,tag,true,&Default);
  }
  template<class T> void getParameter(T& x,const char *tag){
    getParameterBase(x,tag,false,(T*)NULL);
  }

  class BagItem{
  public:
    BagItem *next; 
    void *value;
    int type;
    BagItem(){ next=0; value=0; type=-1; }
    virtual ~BagItem(){}
    virtual void write(std::ostream &os) = 0;
    virtual void free() = 0;
    virtual void set(void *x) = 0;
    //virtual get(void *x) = 0;
  };
  

  struct BagItemType{ char *tag; const char *typeidname; uint n; };
  extern Mem<BagItemType> BagItemTypes;

  template<class T> class BagTypedItem:public BagItem{
  public:
    virtual void write(std::ostream &os){
      if(!value) return;
      if(!BagItemTypes(type).n){
        os <<*((T*)value);
      }else{
        T *t=(T*)value;
        uint n=BagItemTypes(type).n;
        for(uint i=0;i<n;i++,t++) os <<' ' <<*t;
      }
    }
    virtual void free(){
      if(!value) return;
      if(!BagItemTypes(type).n){
        delete ((T*)value);
      }else{
        delete[] ((T*)value);
      }
      value=0;
    }
    virtual void set(void *_x){
      if(!_x){ value=0; return; }
      uint n=BagItemTypes(type).n;
      T *t=(T*)value;
      T *x=(T*)_x;
      if(value) free();
      if(!n){
        value = new T;
        t=(T*)value;
        *t=*x;
      }else{
        value = new T[n];
        t=(T*)value;
        for(uint i=0;i<n;i++) t[i]=x[i];
      }
    }
  };

#ifndef MT_MSVC
  //! add an arbitrary-type item to the bag
  template<class T> void Bag::add(const char *tag,T *x,uint n){
    BagItem *a=new BagTypedItem<T>;
    a->type=getType(tag,typeid(T).name(),n);
    if(a->type==-1){ delete a; return; }
    a->set(x);
    a->next=first;
    first=a;
  }
  
  //! get the item with tag from the bag
  template<class T> BagItem* Bag::get(const char *tag,T *&x,uint n){
    int i;
    BagItem *a;
    i=getType(tag,typeid(T).name(),n);
    for(a=first;a;a=a->next) if(a->type==i) break;
    if(a) x=(T*)a->value; else x=NULL;
    return a;
  }
#endif
}


